/* CS 472/672 Assignment #3 - main.c file */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "rcm.h" 
#include "mindeg.h" 
#include "matrhs.h"
#include "itsolver.h"
#include "ilu.h"


int main(int argc, char **args)
{ 
/*
   Variable definitions:

      Matrix:
      -------
      ia,ja,a    usual ia-ja arrays
      b          right hand side (noisy image)
      x          solution 

      n          the grid size is nxn
      N          number of unknowns

      Iterative methods:
      ------------------
      solver     =0 Jacobi
                 =1 Gauss-Seidel
                 =2 SOR
                 =3 CG
                 =4 PCG

      ILU:
      ----
      order      ordering vector: order( new_order ) = old_order
      invord     inverse of lorder:  invord( order(i) ) = i
      norder     =0 original ordering
                 =1 RCM ordering
                 =2 Min degree ordering
      ILU        ILU matrix in sparse row format

      Convergence of iterative solvers:
      --------------------------------
      tol        convergence tolerance
      maxiter    max number of (inner) iterations
      iter       number of (inner) iterations
      res        residual norm

      Denoising parameters:
      ---------------------
      alpha      parameter for image denoising
                 =4e-2    n=16
                 =3e-2    n=32
                 =1.5e-2  n=64
                 =1.2e-2  n=128

      K          number of denoising (outer) iterations
      imagefile  output file for the denoised image
*/

   char    *usageMessage =
           "Usage: ./programOpt (programDebug) n solver \n"
           "       n = grid size in each dimension for the image.\n"
           "       solver = 0 for Jacobi, 1 for GS, 2 for SOR, 3 for CG, 4 for PCG.\n\n"
           "       if solver=2 (i.e. SOR), \n"
           "           ./programOpt (programDebug) n solver omega \n"
           "           omega = relaxation parameter (default=1).\n\n"
           "       if solver=4 (i.e. PCG), \n"
           "           ./programOpt (programDebug) n solver norder \n"
           "           norder = 0 for original ordering, 1 for RCM, 2 for min deg (default=0).\n";
   int     n, N, i, k, K, iter, totalIter, maxiter, solver, norder;
   int     *ia, *ja, *order, *invord;
   double  *a, *x, *b, alpha, tol, res, omega;
   FILE    *imagefile;
   struct  sparse_row *ILU;
   clock_t afterInitial, afterAnalyze, afterFactor, afterSolve, startIter, totalSolve;


   /* Set up parameters. */
   n = 16;
   solver = 0;
   omega = 1.0;
   norder = 0;

   if (argc==1) {printf("%s",usageMessage); exit(0);}
   if (argc>1) n = atoi(args[1]);
   if (argc>2) solver = atoi(args[2]);
   if (solver==2) if (argc>3) omega = atof(args[3]);
   if (solver==4) if (argc>3) norder = atoi(args[3]);
   if (argc>4) {printf("%s",usageMessage); exit(0);}

   N = n*n;
   K = 10;
   maxiter = 100000;
   tol = 1e-2;
   alpha = (n==16) ? 4e-2 : (n==32) ? 3e-2 : (n==64) ? 1.5e-2 : (n==128) ? 1.2e-2 : 1e-2;
   totalIter = 0;
   totalSolve = 0;

   printf("Solving the Nonlinear Image Denoising Problem\n");
   printf("Grid size = %d x %d\n",n,n);
   printf("Solver = %s\n",((solver==0) ? "Jacobi" : (solver==1) ? "Gauss-Seidel" : (solver==2) ? "SOR" : (solver==3) ? "Conjugate gradient" : (solver==4) ? "Preconditioned conjugate gradient" : "Error"));
   if (solver==2) printf("Omega = %f\n",omega);
   if (solver==4) printf("Ordering method = %s\n",((norder==0) ? "Natural" : (norder==1) ? "RCM" : (norder==2) ?
 "Minimum degree" : "Error"));
   printf("Denoising parameter, alpha = %f\n",alpha);
   printf("Maximum (inner) linear iteration = %d\n",maxiter);
   printf("Maximum (outer) nonlinear iteration = %d\n",K);
   printf("Tolerance for convergence test = %f\n\n",tol);


   /* Allocate space for arrays ia, ja, a, x, and b.              */
   /* Allocate space for order and inverse order vectors.         */
   /* Allocate array of sparse_row structures for the ILU matrix. */
   ia = (int *) malloc((N+1)*sizeof(int));
   ja = (int *) malloc(5*N*sizeof(int));
   a = (double *) malloc(5*N*sizeof(double));
   x = (double *) malloc(N*sizeof(double));
   b = (double *) malloc(N*sizeof(double));

   if (solver==4) {
     order = (int *) malloc(N*sizeof(int));
     invord = (int *) malloc(N*sizeof(int));
     ILU = (struct sparse_row *) malloc(N*sizeof(struct sparse_row));
   }


   /* Set up initial guess, RHS (noisy image), and initial matrix */
   noisyimg(n, x, b);
   imgmatrix(n, alpha, x, ia, ja, a);

   afterInitial = clock();


   /* Initial setup for ILU: ordering and analyze phase */
   if (solver==4) {

     /*  Ordering and inverse ordering. */
     if (norder==0) for (i=0; i<N; i++) order[i] = i;
     if (norder==1) rcm(ia, ja, N, order);
     if (norder==2) mindeg(ia, ja, N, order);
     for (i=0; i<N; i++) invord[order[i]] = i;

     /* Symbolic incomplete factorization */
     ilusym(N, ia, ja, order, invord, ILU);

     afterAnalyze = clock();
     printf("Time for ordering and anlyze phase = %f\n",
            ((double) (afterAnalyze-afterInitial)/(double) CLOCKS_PER_SEC));
   }


   /* Denoising (outer) iteration */
   for (k=0; k<K; k++) {

      startIter = clock();
      printf("Starting (Outer) Iteration %d:\n",k+1);

      /* Calling the iterative solver */
      if (solver==0) Jacobi(N, ia, ja, a, b, x, maxiter, tol, &iter, &res);
      if (solver==1) GS(N, ia, ja, a, b, x, maxiter, tol, &iter, &res);
      if (solver==2) SOR(N, ia, ja, a, b, x, omega, maxiter, tol, &iter, &res);
      if (solver==3) CG(N, ia, ja, a, b, x, maxiter, tol, &iter, &res);

      if (solver==4) {

         /* Incomplete LU factorization */
         ilufac(N, ia, ja, a, order, invord, ILU);

         afterFactor = clock();
         printf("    Time for incomplete factorization = %f\n",
                ((double) (afterFactor - startIter)/(double) CLOCKS_PER_SEC));

         PCG(N, ia, ja, a, b, x, order, ILU, maxiter, tol, &iter, &res);
      }

      /* Show timing and convergence of the iterative solver */
      afterSolve  = clock();
      printf("    Time for solver = %f\n",
             ((double) (afterSolve - startIter)/(double) CLOCKS_PER_SEC));
      printf("    Number of iterations = %d\n",iter);
      if (iter==maxiter) {
         printf("The iterative method does not converge in %d iterations.\n",maxiter);
         printf("Residual norm = %e.\n",res);
         exit(0);
      }
      printf("    Residual norm = %e\n\n",res);

      totalSolve += afterSolve - startIter;
      totalIter += iter;

      /* Update the matrix using the current values of x (the last solution) */
      imgmatrix(n, alpha, x, ia, ja, a);

   }


   /* Report the total time and number of iterations */
   printf("Total solve time = %f\n",
          ((double) totalSolve/(double) CLOCKS_PER_SEC));
   printf("Average number of iterations = %f\n",totalIter*1.0/K);


   /* Save the denoised image to imagefile */
   imagefile=fopen("img.m","w");
   for (i=0; i<N; i++) fprintf(imagefile,"%25.15e\n",x[i]);
   fclose(imagefile);


   /* Free memory  */
   free(ia); free(ja); free(a); free(x); free(b);
   if (solver==4) {free(order); free(invord); clean(N, ILU);}

   return 0;
}
